"""Types for solc."""

from typing import TypedDict

from web3.types import ABI


class SolcContract(TypedDict):
    """Foundry"""

    abi: ABI
    bin: str
    metadata: str


class SolcJson(TypedDict):
    """Foundry"""

    contracts: dict[str, SolcContract]
    version: str



"""Dataclasses for all structs in the OverloadedMethods contract."""
# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ
# contracts have PascalCase names
# pylint: disable=invalid-name
# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import
# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return
from __future__ import annotations



"""Utilities for templating."""
import os

from jinja2 import Environment, FileSystemLoader


def get_jinja_env() -> Environment:
    """Returns the jinja environment."""
    script_dir = os.path.dirname(os.path.abspath(__file__))
    # Construct the path to your templates directory.
    templates_dir = os.path.join(script_dir, "../templates")
    env = Environment(loader=FileSystemLoader(templates_dir))
    return env



"""Functions to render Python files from an abi usng a jinja2 template."""
from __future__ import annotations

from pathlib import Path
from typing import Any, NamedTuple

from web3.types import ABI

from pypechain.utilities.abi import (
    get_abi_items,
    get_input_names,
    get_input_names_and_values,
    get_input_types,
    get_output_names,
    get_output_names_and_values,
    get_output_types,
    get_structs_for_abi,
    is_abi_constructor,
    is_abi_function,
    load_abi_from_file,
)
from pypechain.utilities.format import capitalize_first_letter_only
from pypechain.utilities.templates import get_jinja_env
from pypechain.utilities.types import FunctionData, SignatureData, gather_matching_types


def render_contract_file(contract_name: str, abi_file_path: Path) -> str:
    """Returns the serialized code of the contract file to be generated.

    Arguments
    ---------
    contract_template : Template
        A jinja template containging types for all structs within an abi.
    abi_file_path : Path
        The path to the abi file to parse.

    Returns
    -------
    str
        A serialized python file.
    """
    env = get_jinja_env()
    templates = get_templates_for_contract_file(env)

    abi, bytecode = load_abi_from_file(abi_file_path)
    function_datas, constructor_data = get_function_datas(abi)
    has_overloading = any(len(function_data["signature_datas"]) > 1 for function_data in function_datas.values())
    has_bytecode = bool(bytecode)

    structs_for_abi = get_structs_for_abi(abi)
    structs_used = gather_matching_types(list(function_datas.values()), list(structs_for_abi.keys()))

    functions_block = templates.functions_template.render(
        abi=abi,
        has_overloading=has_overloading,
        contract_name=contract_name,
        functions=function_datas,
        # TODO: use this data to add a typed constructor
        constructor=constructor_data,
    )

    abi_block = templates.abi_template.render(
        abi=abi,
        bytecode=bytecode,
        contract_name=contract_name,
    )

    contract_block = templates.contract_template.render(
        has_bytecode=has_bytecode,
        contract_name=contract_name,
        functions=function_datas,
    )

    # Render the template
    return templates.base_template.render(
        contract_name=contract_name,
        structs_used=structs_used,
        has_overloading=has_overloading,
        has_bytecode=has_bytecode,
        functions_block=functions_block,
        abi_block=abi_block,
        contract_block=contract_block,
        # TODO: use this data to add a typed constructor
        # constructor_data=constructor_data,
    )


class ContractTemplates(NamedTuple):
    """Templates for the generated contract file."""

    base_template: Any
    functions_template: Any
    abi_template: Any
    contract_template: Any


def get_templates_for_contract_file(env):
    """Templates for the generated contract file."""
    return ContractTemplates(
        base_template=env.get_template("contract.py/base.py.jinja2"),
        functions_template=env.get_template("contract.py/functions.py.jinja2"),
        abi_template=env.get_template("contract.py/abi.py.jinja2"),
        contract_template=env.get_template("contract.py/contract.py.jinja2"),
    )


def get_function_datas(abi: ABI) -> tuple[dict[str, FunctionData], SignatureData | None]:
    """_summary_

    Arguments
    ---------
    abi : ABI
        An application boundary interface for smart contract in json format.

    Returns
    -------
    tuple[dict[str, FunctionData], SignatureData | None]
        A tuple where the first value is a dictionary of FunctionData's keyed by function name and
        the second value is SignatureData for the constructor.
    """
    function_datas: dict[str, FunctionData] = {}
    constructor_data: SignatureData | None = None
    for abi_function in get_abi_items(abi):
        if is_abi_function(abi_function):
            # hanndle constructor
            if is_abi_constructor(abi_function):
                constructor_data = {
                    "input_names_and_types": get_input_names_and_values(abi_function),
                    "input_names": get_input_names(abi_function),
                    "input_types": get_input_types(abi_function),
                    "outputs": get_output_names(abi_function),
                    "output_types": get_output_names_and_values(abi_function),
                }

            # handle all other functions
            else:
                name = abi_function.get("name", "")
                signature_data: SignatureData = {
                    "input_names_and_types": get_input_names_and_values(abi_function),
                    "input_names": get_input_names(abi_function),
                    "input_types": get_input_types(abi_function),
                    "outputs": get_output_names(abi_function),
                    "output_types": get_output_types(abi_function),
                }
                function_data: FunctionData = {
                    "name": name,
                    "capitalized_name": capitalize_first_letter_only(name),
                    "signature_datas": [signature_data],
                }
                if not function_datas.get(name):
                    function_datas[name] = function_data
                else:
                    function_datas[name]["signature_datas"].append(signature_data)
    return function_datas, constructor_data



"""Tests for overloading methods."""
from __future__ import annotations

import os

import pytest
from web3.exceptions import Web3ValidationError

from pypechain.test.overloading.types.OverloadedMethodsContract import OverloadedMethodsContract

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestOverloading:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_overloading(self, w3):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        # TODO: add the factory to the constructor so we can do:
        # deployed_contract = MyContract(w3=w3, address=address)
        # or:
        # ContractDeployer =  MyContract(w3)
        # deployed_contract = ContractDeployer.deploy(arg1, arg2)
        deployer = OverloadedMethodsContract.factory(w3=w3)

        # TODO: put this into a deploy method and add the address automatically
        tx_hash = deployer.constructor().transact({"from": w3.eth.accounts[0]})
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        deployed_contract = deployer(address=tx_receipt.contractAddress)

        s = "test string"
        x = 1
        y = 2

        # TODO:
        result = deployed_contract.functions.doSomething(s).call()
        assert result == "test string"

        result = deployed_contract.functions.doSomething(x).call()
        assert result == 1 * 2

        result = deployed_contract.functions.doSomething(x, y).call()
        assert result == 1 + 2

        with pytest.raises(Web3ValidationError) as err:
            result = deployed_contract.functions.doSomething(x, y, s).call()
        assert "Could not identify the intended function with name `doSomething`" in str(err.value)



"""Tests for rendering the contract file."""
import json
import os

from web3.types import ABI

from pypechain.render.contract import get_function_datas
from pypechain.utilities.templates import get_jinja_env

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestOverloading:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_overloading(self, snapshot):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        env = get_jinja_env()
        functions_template = env.get_template("contract.py/functions.py.jinja2")

        # TODO: add return types to function calls

        abi_str = """
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                },
                {
                    "constant": true,
                    "inputs": [{"name": "who", "type": "address"}],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
        """

        abi: ABI = json.loads(abi_str)

        function_datas, constructor_data = get_function_datas(abi)
        has_overloading = any(len(function_data["signature_datas"]) > 1 for function_data in function_datas.values())
        contract_name = "Overloaded"

        functions_block = functions_template.render(
            abi=abi,
            contract_name=contract_name,
            functions=function_datas,
            # TODO: use this data to add a typed constructor
            constructor=constructor_data,
        )
        assert has_overloading is True

        snapshot.snapshot_dir = "snapshots"  # This line is optional.
        snapshot.assert_match(functions_block, "expected_overloading.py")

    def test_notoverloading(self, snapshot):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        env = get_jinja_env()
        functions_template = env.get_template("contract.py/functions.py.jinja2")

        # TODO: add return types to function calls

        # different names, should NOT be overloaded
        abi_str = """
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                },
                {
                    "constant": true,
                    "inputs": [{"name": "who", "type": "address"}],
                    "name": "balanceOfWho",
                    "outputs": [{"name": "", "type": "bool"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
        """

        abi: ABI = json.loads(abi_str)

        function_datas, constructor_data = get_function_datas(abi)
        has_overloading = any(len(function_data["signature_datas"]) > 1 for function_data in function_datas.values())
        contract_name = "Overloaded"

        functions_block = functions_template.render(
            abi=abi,
            contract_name=contract_name,
            functions=function_datas,
            # TODO: use this data to add a typed constructor
            constructor=constructor_data,
        )
        assert has_overloading is False

        snapshot.snapshot_dir = "snapshots"
        snapshot.assert_match(functions_block, "expected_not_overloading.py")
